from inc_noesis import *

import noesis
import rapi
import os
import hashlib
import json
import sys


from ctypes import *
import noewin

import nrfile
import nrtools
import nrdump


# Default options
g_importOptions = {
'importType': 'local',             # local/world
'texMode':    'ByAttributeName', 
'attribName': 'TEXCOORD',
'uvIdx':      0,
'uCompIdx':   0,
'vCompIdx':   1,
'normalVecMode': 'Disabled',
'projMat': [[1.8106601717798212, 0.0, 0.0, 0.0], [0.0, 2.414213562373095, 0.0, 0.0], [0.0, 0.0, 1.000010000100001, 1.0], [0.0, 0.0, -0.01000010000100001, 0.0]],  # 45.0/1024/768/0.01/1000
'pos_x_attrIdx': 0,
'pos_y_attrIdx': 0,
'pos_z_attrIdx': 0,
'pos_x_compIdx': 0,
'pos_y_compIdx': 1,
'pos_z_compIdx': 2
}

def getImportOptions():
    global g_importOptions
    return g_importOptions

def setImportOptions(opt):
    global g_importOptions
    g_importOptions = opt


def writeLog(s):
    print(s)


def isNoesisCanSetSelectDir():
    return hasattr(noesis, "setSelectedDirectory")


def setRipperDir():
    if isNoesisCanSetSelectDir():
        ripDir = nrtools.getNrOutputDir()
        if ripDir:
            print("Setting the ripper directory {:s}".format(ripDir))
            noesis.setDataViewerSetting("Hidden items in shellview", "1")
            noesis.setSelectedDirectory(ripDir)


class NoesisImporter(object):
    def importPostVs(self, filename, options, mdlList):
        #print(options)

        nr = nrfile.NRFile()
        if not nr.parse(filename):
            errMsg = "Ninja Ripper file parsing failed: " + nr.getErrorString()
            writeLog(errMsg)
            return retVal(False, errMsg)

        writeLog("Loading: {:s}".format(filename))

        fileDirectory = os.path.dirname(os.path.abspath(filename))
        
        createdMeshsCount = 0

        for meshIdx in range(0, nr.getMeshCount()):
            nrmesh = nr.getMesh(meshIdx)

            if nrfile.ShaderStage.PreVs == nrmesh.getShaderStage():
                #writeLog("Skipping PreVs")
                continue

            vert  = nrmesh.getVertexes(0)
            if None == vert:
                writeLog("vertexes == Null")
                continue
            
            indx  = nrmesh.getIndexes(0)
            if None == indx:
                writeLog("indexes == Null")
                continue
            
            if indx.getIndexTopology() != nrfile.IndexTopology.TriangleList:
                writeLog("IndexTopology != TriangleList  (Topology={:s})".format(nrdump.topologyToStr(indx.getIndexTopology())))
                continue
        
            vatrs = nrmesh.getVertexAttributes(0)
            if None == vatrs:
                writeLog("vertexAttribs == Null")
                continue
            
            meshName = os.path.basename(filename)

            # Extra vatrs/vertexes
            vert1  = nrmesh.getVertexes(1)
            vatrs1 = nrmesh.getVertexAttributes(1)
            vertexData1 = None
            if vert1:
                vertexData1 = vert1.read()

            extraVertexDataOk = True
            if (not vert1) or (not vertexData1) or (not vatrs1):
                extraVertexDataOk = False

            if extraVertexDataOk:
                if vert.getVertexCount() != vert1.getVertexCount():
                    extraVertexDataOk = False
                    writeLog("Vertex count != Extra vertex count {:d} != {:d}".format(vert.getVertexCount(), vert1.getVertexCount()))

            # Get Textures object (TXTR tag in rip file)
            #  textures = texture files list
            # Material create
            material = None
            matList = []
            texList = []
            textures = nrmesh.getTextures()
            if textures:
                if textures.getTexturesCount() > 0:
                    tex0 = textures.getTexture(0)
                    texName = fileDirectory + "\\" + tex0.fileName
                    textureObj = rapi.loadExternalTex(texName)
                    if textureObj:
                        material = NoeMaterial(texName, texName)
                        material.setTexture(texName)
                        matList.append(material)
                        texList.append(textureObj)
                        rapi.rpgSetMaterial(texName)

            # Attr 0 is POSITION
            if vatrs.getAttr(0).compCount != 4:
                writeLog("Post vs position not 4 component")
                continue

            vertexData = vert.read()
            if None == vertexData:
                writeLog("vertexData == Null")
                continue

            #projMat = nrtools.perspMatFovLH(options.fov, options.width, options.height, options.nearDist, options.farDist)
            positions3 = nrtools.restorePositionAsBytes(vert, vertexData, vatrs.getAttr(0), options['projMat'])
            if None == positions3:
                writeLog("positions3 == Null")
                continue

            rapi.rpgBindPositionBuffer(positions3, noesis.RPGEODATA_FLOAT, 12)
            
            addUvCoords(options, vert, vertexData, vatrs, extraVertexDataOk, vert1, vertexData1, vatrs1)
            addNormals(options, vert, vertexData, vatrs)

            triangles = indx.read()
            rapi.rpgCommitTriangles(triangles, noesis.RPGEODATA_UINT, indx.getIndexCount(), noesis.RPGEO_TRIANGLE, 1)

            rapi.rpgClearBufferBinds() #reset in case a subsequent mesh doesn't have the same components

            mdl = rapi.rpgConstructModel()

            if material:
                mdl.setModelMaterials(NoeModelMaterials(texList, matList))
            mdlList.append(mdl)

        return 1




    def importPreVs(self, filename, options, mdlList):
        #print(options)

        nr = nrfile.NRFile()
        if not nr.parse(filename):
            errMsg = "Ninja Ripper file parsing failed: " + nr.getErrorString()
            writeLog(errMsg)
            return retVal(False, errMsg)

        writeLog("PreVS Loading: {:s}".format(filename))
        fileDirectory = os.path.dirname(os.path.abspath(filename))
        for meshIdx in range(0, nr.getMeshCount()):
            nrmesh = nr.getMesh(meshIdx)

            if nrfile.ShaderStage.PreVs != nrmesh.getShaderStage():
                #writeLog("Skipping PreVs")
                continue

            vert  = nrmesh.getVertexes(0)
            if None == vert:
                writeLog("vertexes == Null")
                continue
            
            indx  = nrmesh.getIndexes(0)
            if None == indx:
                writeLog("indexes == Null")
                continue
            
            if indx.getIndexTopology() != nrfile.IndexTopology.TriangleList:
                writeLog("IndexTopology != TriangleList  (Topology={:s})".format(nrdump.topologyToStr(indx.getIndexTopology())))
                continue
        
            vatrs = nrmesh.getVertexAttributes(0)
            if None == vatrs:
                writeLog("vertexAttribs == Null")
                continue
            
            meshName = os.path.basename(filename)

            # Extra vatrs/vertexes
            vert1  = nrmesh.getVertexes(1)
            vatrs1 = nrmesh.getVertexAttributes(1)
            vertexData1 = None
            if vert1:
                vertexData1 = vert1.read()

            extraVertexDataOk = True
            if (not vert1) or (not vertexData1) or (not vatrs1):
                extraVertexDataOk = False

            if extraVertexDataOk:
                if vert.getVertexCount() != vert1.getVertexCount():
                    extraVertexDataOk = False
                    writeLog("Vertex count != Extra vertex count {:d} != {:d}".format(vert.getVertexCount(), vert1.getVertexCount()))

            # Get Textures object (TXTR tag in rip file)
            #  textures = texture files list
            # Material create
            material = None
            matList = []
            texList = []
            textures = nrmesh.getTextures()
            # With textures some meshs don't displayed
            if textures:
                if textures.getTexturesCount() > 0:
                    tex0 = textures.getTexture(0)
                    texName = fileDirectory + "\\" + tex0.fileName
                    textureObj = rapi.loadExternalTex(texName)
                    if textureObj:
                        material = NoeMaterial(texName, texName)
                        material.setTexture(texName)
                        matList.append(material)
                        texList.append(textureObj)
                        rapi.rpgSetMaterial(texName)

            vertexData = vert.read()
            if None == vertexData:
                writeLog("vertexData == Null")
                continue

            positions3 = loadPosition3(options, vert, vertexData, vatrs)
            if None == positions3:
                writeLog("positions3 == Null")
                continue

            rapi.rpgBindPositionBuffer(positions3, noesis.RPGEODATA_FLOAT, 12)
            
            addUvCoords(options, vert, vertexData, vatrs, extraVertexDataOk, vert1, vertexData1, vatrs1)
            addNormals(options, vert, vertexData, vatrs)

            triangles = indx.read()
            rapi.rpgCommitTriangles(triangles, noesis.RPGEODATA_UINT, indx.getIndexCount(), noesis.RPGEO_TRIANGLE, 1)

            rapi.rpgClearBufferBinds() #reset in case a subsequent mesh doesn't have the same components

            mdl = rapi.rpgConstructModel()

            if material:
                mdl.setModelMaterials(NoeModelMaterials(texList, matList))

            mdlList.append(mdl)
        return 1




def addUvCoords(options, vert, vertexData, vatrs, extraVertexDataOk, vert1, vertexData1, vatrs1):
    if options.get('texMode'):
        if (options['texMode'] == 'ByAttributeName'):
            createTexCoords(options, vert, vertexData, vatrs)
        elif (options['texMode'] == 'ByAttributeIndex'):
            createTexCoordsScatter(options, vert, vertexData, vatrs)
        elif (options['texMode'] == 'Extra_ByAttributeName') and extraVertexDataOk:
            createTexCoords(options, vert1, vertexData1, vatrs1)
        elif (options['texMode'] == 'Extra_ByAttributeIndex') and extraVertexDataOk:
            createTexCoordsScatter(options, vert1, vertexData1, vatrs1)


def createTexCoordsScatter(options, vert, vertexData, vatrs):
    texUattr  = options['uAttribIdx']  # [attrIdx]
    texUidx   = options['uCompIdx']    # [compIdx]

    texVattr  = options['vAttribIdx']  # [attrIdx]
    texVidx   = options['vCompIdx']    # [compIdx]

    uvs = nrtools.unpackVertexComponentVaAsBytes(vert, vertexData, vatrs, [[texUattr, texUidx], [texVattr,texVidx]])
    if None == uvs:
        writeLog('ERROR: UV loading failed')
        return

    rapi.rpgBindUV1Buffer(uvs, noesis.RPGEODATA_FLOAT, 8)


def createTexCoords(options, vert, vertexData, vatrs):
    texCoordAttribName  = options['attribName']  # Default TEXCOORD
    uvIdx               = options['uvIdx']
    texUvComponentU_Idx = options['uCompIdx']
    texUvComponentV_Idx = options['vCompIdx']

    texcoordVaList = vatrs.findSemantic(texCoordAttribName)
    if 0 == len(texcoordVaList):
        writeLog('ERROR: No attributes with name {:s} found'.format(texCoordAttribName))
        return

    if uvIdx >= len(texcoordVaList):
        writeLog('ERROR: UV index={:d} >= Size={:d}'.format(uvIdx, len(texcoordVaList)))
        return

    tcVa = texcoordVaList[uvIdx]
    if tcVa.getComponentsCount() < 2:
        writeLog("ERROR: UV components < 2")
        return

    if texUvComponentU_Idx > tcVa.getComponentsCount():
        writeLog('ERROR: texUvComponentU_Idx {:d} > tcVa.getComponentsCount() {:d}'.format(texUvComponentU_Idx, tcVa.getComponentsCount()))
        return

    if texUvComponentV_Idx > tcVa.getComponentsCount():
        writeLog('ERROR: texUvComponentV_Idx {:d} > tcVa.getComponentsCount() {:d}'.format(texUvComponentV_Idx, tcVa.getComponentsCount()))
        return

    uvs = nrtools.unpackVertexVaCompAsBytes(vert, vertexData, tcVa, [texUvComponentU_Idx, texUvComponentV_Idx])
    if None == uvs:
        writeLog('ERROR: UV loading failed')
        return

    rapi.rpgBindUV1Buffer(uvs, noesis.RPGEODATA_FLOAT, 8)



def addNormalVecAttrib(options, vert, vertexData, vatrs):
    xAttrIdx = options['norm_x_attrIdx']
    xCompIdx = options['norm_x_compIdx']

    yAttrIdx = options['norm_y_attrIdx']
    yCompIdx = options['norm_y_compIdx']

    zAttrIdx = options['norm_z_attrIdx']
    zCompIdx = options['norm_z_compIdx']

    uvs = nrtools.unpackVertexComponentVaAsBytes(vert, vertexData, vatrs, [[xAttrIdx, xCompIdx], [yAttrIdx,yCompIdx], [zAttrIdx,zCompIdx]])
    if None == uvs:
        writeLog('ERROR: Normals loading failed')
        return

    rapi.rpgBindNormalBuffer(uvs, noesis.RPGEODATA_FLOAT, 12)


def addNormals(options, vert, vertexData, vatrs):
    if options.get('normalVecMode'):
        if options['normalVecMode'] == 'Manual':
            addNormalVecAttrib(options, vert, vertexData, vatrs)


def loadPosition3(options, vert, vertexData, vatrs):
    xAttrIdx = options['pos_x_attrIdx']
    xCompIdx = options['pos_x_compIdx']

    yAttrIdx = options['pos_y_attrIdx']
    yCompIdx = options['pos_y_compIdx']

    zAttrIdx = options['pos_z_attrIdx']
    zCompIdx = options['pos_z_compIdx']

    #writeLog("--->{:d}/{:d} {:d}/{:d}  {:d}/{:d}".format(xAttrIdx, xCompIdx, yAttrIdx, yCompIdx, zAttrIdx,zCompIdx))

    pos3 = nrtools.unpackVertexComponentVaAsBytes(vert, vertexData, vatrs, [[xAttrIdx, xCompIdx], [yAttrIdx,yCompIdx], [zAttrIdx,zCompIdx]])
    if None == pos3:
        writeLog('ERROR: Position loading failed')
        return None

    return pos3
